-- isolines and isobands

local iso = {}


--[[
The iso.createNoise function takes several optional parameters: 
noiseFunction (defaulting to love.math.noise), 
shiftX, shiftY, scaleX, scaleY, and angle. 
These parameters control the transformation 
and scaling applied to the input coordinates 
before passing them to the noiseFunction.

The resulting noise function is assigned to iso.noise 
and can be called with x and y coordinates 
to retrieve the noise value.

You can use this iso.noise function to generate 
noise values based on the provided parameters.
]]

function iso.createNoise(noiseFunction, shiftX, shiftY, scaleX, scaleY, angle)
  noiseFunction = noiseFunction or love.math.noise
  shiftX = shiftX or 0
  shiftY = shiftY or 0
  scaleX = scaleX or 1
  scaleY = scaleY or 1
  angle = angle or 0
  
  return function(x, y)
    x, y = x*math.cos(angle)-y*math.sin(angle), x*math.sin(angle)+y*math.cos(angle)
    x = (x + shiftX) * scaleX
    y = (y + shiftY) * scaleY
		local value = noiseFunction(x, y)
		value = math.floor(value*5+0.5)/5
--		print (x, y, value)
    return value
  end
end


--[[
The function iterates over each row and column 
of the grid and calls iso.noise 
to generate a noise value for each position. 
The resulting values are stored in the grid table.
]]

function iso.createNoiseGrid (width, height, noise)
	local xMin, xMax = 1, width+1
	local yMin, yMax = 1, height+1
  local noiseGrid = {}
  for y = yMin, yMax do
		noiseGrid[y] = {}
    for x = xMin, xMax do
      local value = noise (x, y)
      noiseGrid[y][x] = value
    end
  end
	iso.xMin, iso.xMax = xMin, xMax
	iso.yMin, iso.yMax = yMin, yMax
  return noiseGrid
end


--[[
The function calculates the interpolation parameter t 
and then checks if it is less than 0 or greater than 1. 
If t is less than 0, it returns the coordinates 
of the first point along with false to indicate 
that no interpolation was performed. If t is greater than 1, 
it returns the coordinates of the second point along with false. 
Otherwise, if t is within the range [0, 1], it performs 
linear interpolation to determine the x and y coordinates 
of the interpolated point and returns them along with true 
to indicate successful interpolation.
]]
local function interpolatePoint(x1, y1, v1, x2, y2, v2, isovalue)
  -- Calculate the interpolation parameter t
  local t = (isovalue - v1) / (v2 - v1)

  -- Check if t is less than 0
  if t < 0 then
    -- Return the coordinates of the first point and false to indicate no interpolation
    return x1, y1, false
  end

  -- Check if t is greater than 1
  if t > 1 then
    -- Return the coordinates of the second point and false to indicate no interpolation
    return x2, y2, false
  end

  -- Perform linear interpolation to determine the x and y coordinates of the interpolated point
  local x = x1 + t * (x2 - x1)
  local y = y1 + t * (y2 - y1)

  -- Return the interpolated x and y coordinates along with true to indicate successful interpolation
  return x, y, true
end


--local function interpolateAndInsertPoint(x1, y1, v1, x2, y2, v2, isovalue, line)
--  local px, py, success = interpolatePoint(x1, y1, v1, x2, y2, v2, isovalue)
--  if success then
--    table.insert(line, px)
--    table.insert(line, py)
--  end
--end

local function findIsoLine(x, y, isovalue, scale, noiseGrid)
	scale = scale or 1
--	print ('scale:', scale)
  noiseGrid = noiseGrid or iso.noiseGrid
  local xs = {x, x+1, x+1, x}
  local ys = {y, y, y+1, y+1}
  local vs = {noiseGrid[y][x], noiseGrid[y][x+1], noiseGrid[y+1][x+1], noiseGrid[y+1][x]}
  local line = {}
  for i = 1, 4 do
    local j = i % 4 + 1
    local x1, y1, v1 = xs[i], ys[i], vs[i]
    local x2, y2, v2 = xs[j], ys[j], vs[j]
    if math.min(v1, v2) < isovalue and math.max(v1, v2) >= isovalue then
			local px, py, success = interpolatePoint(x1, y1, v1, x2, y2, v2, isovalue)
			if success then
				table.insert(line, (px-1)*scale)
				table.insert(line, (py-1)*scale)
			end
    end
  end

  if #line == 8 then
    if math.abs(vs[1] + vs[3] - 2 * isovalue) > math.abs(vs[2] + vs[4] - 2 * isovalue) then
      return {line[1], line[2], line[3], line[4]}, {line[5], line[6], line[7], line[8]}
    else
      return {line[1], line[2], line[7], line[8]}, {line[5], line[6], line[3], line[4]}
    end
  elseif #line == 4 then
    return line
  end
end

function iso.createIsolines (isovalue, scale, noiseGrid)
--	print (scale)
	local isolines = {} -- list of lines for isovalues
	for y = iso.yMin, iso.yMax-1 do
		for x = iso.xMin, iso.xMax-1 do
			local line1, line2 = findIsoLine(x, y, isovalue, scale, noiseGrid)
			table.insert (isolines, line1)
			table.insert (isolines, line2)
		end
	end
--	if not iso.isolines then iso.isolines = {} end
--	iso.isolines[isovalue] = isolines
	return isolines
end



local function createNoisePoints (isovalueMin, isovalueMax, r, g, b)
	local noiseMap = iso.noiseGrid
	local noisePoints = {}
	for y = iso.yMin, iso.yMax do
		for x = iso.xMin, iso.xMax do
			local t = noiseMap[y][x]
			if isovalueMin < t and t <= isovalueMax then
				table.insert(noisePoints, {x, y, r, g, b ,1})
			end
		end
	end
	
	if not iso.noisePoints then iso.noisePoints = {} end
	iso.noisePoints[isovalueMin..'-'..isovalueMax] = noisePoints
	return noisePoints
end




function createIsopolygon (noiseMap, x, y, isovalue1, isovalue2)
	local corners = {{x,y},{x+1,y},{x+1,y+1},{x,y+1}}
	
	local values = {}
	for i = 1, 4 do
		if noiseMap[corners[i][2] ] and noiseMap[corners[i][2] ][corners[i][1] ] then
			local vx, vy = corners[i][1], corners[i][2]
			values[i] = noiseMap[vy][vx]
		else
			return
		end
	end	
	
	local points = {}
	local edges = {}
	
	if    values[1] > math.min (isovalue1, isovalue2)
		and values[1] < math.max (isovalue1, isovalue2) then 
		-- adding top left vertice
		local px, py = corners[1][1], corners[1][2]
		table.insert(points, px)
		table.insert(points, py)
	end
	
	
	if values[1] < values[2] then
		if math.min(values[1], values[2]) < isovalue1 and math.max(values[1], values[2]) > isovalue1 then
			local px, py = interpolatePoint(corners[1][1], corners[1][2], values[1], corners[2][1], corners[2][2], values[2], isovalue1)
			table.insert(points, px)
			table.insert(points, py)
		end
		if math.min(values[1], values[2]) < isovalue2 and math.max(values[1], values[2]) > isovalue2 then
			local px, py = interpolatePoint(corners[1][1], corners[1][2], values[1], corners[2][1], corners[2][2], values[2], isovalue2)
			table.insert(points, px)
			table.insert(points, py)
		end
	else -- other direction
		if math.min(values[1], values[2]) < isovalue2 and math.max(values[1], values[2]) > isovalue2 then
			local px, py = interpolatePoint(corners[1][1], corners[1][2], values[1], corners[2][1], corners[2][2], values[2], isovalue2)
			table.insert(points, px)
			table.insert(points, py)
		end
		if math.min(values[1], values[2]) < isovalue1 and math.max(values[1], values[2]) > isovalue1 then
			local px, py = interpolatePoint(corners[1][1], corners[1][2], values[1], corners[2][1], corners[2][2], values[2], isovalue1)
			table.insert(points, px)
			table.insert(points, py)
		end
	end
	
	if    values[2] > math.min (isovalue1, isovalue2)
		and values[2] < math.max (isovalue1, isovalue2) then 
		-- adding top right vertice
		local px, py = corners[2][1], corners[2][2]
		table.insert(points, px)
		table.insert(points, py)
	end
	

	if values[2] < values[3] then
		if math.min(values[2], values[3]) < isovalue1 and math.max(values[2], values[3]) > isovalue1 then
			local px, py = interpolatePoint(corners[2][1], corners[2][2], values[2], corners[3][1], corners[3][2], values[3], isovalue1)
			table.insert(points, px)
			table.insert(points, py)
		end
		if math.min(values[2], values[3]) < isovalue2 and math.max(values[2], values[3]) > isovalue2 then
			local px, py = interpolatePoint(corners[2][1], corners[2][2], values[2], corners[3][1], corners[3][2], values[3], isovalue2)
			table.insert(points, px)
			table.insert(points, py)
		end
	else -- other direction
		if math.min(values[2], values[3]) < isovalue2 and math.max(values[2], values[3]) > isovalue2 then
			local px, py = interpolatePoint(corners[2][1], corners[2][2], values[2], corners[3][1], corners[3][2], values[3], isovalue2)
			table.insert(points, px)
			table.insert(points, py)
		end
		if math.min(values[2], values[3]) < isovalue1 and math.max(values[2], values[3]) > isovalue1 then
			local px, py = interpolatePoint(corners[2][1], corners[2][2], values[2], corners[3][1], corners[3][2], values[3], isovalue1)
			table.insert(points, px)
			table.insert(points, py)
		end
	end
	
	if    values[3] > math.min (isovalue1, isovalue2)
		and values[3] < math.max (isovalue1, isovalue2) then 
		-- adding bottom right vertice
		local px, py = corners[3][1], corners[3][2]
		table.insert(points, px)
		table.insert(points, py)
	end
	
	if values[3] < values[4] then
		if math.min(values[4], values[3]) < isovalue1 and math.max(values[4], values[3]) > isovalue1 then
			local px, py = interpolatePoint(corners[4][1], corners[4][2], values[4], corners[3][1], corners[3][2], values[3], isovalue1)
			table.insert(points, px)
			table.insert(points, py)
		end
		if math.min(values[4], values[3]) < isovalue2 and math.max(values[4], values[3]) > isovalue2 then
			local px, py = interpolatePoint(corners[4][1], corners[4][2], values[4], corners[3][1], corners[3][2], values[3], isovalue2)
			table.insert(points, px)
			table.insert(points, py)
		end
	else -- other direction
		if math.min(values[4], values[3]) < isovalue2 and math.max(values[4], values[3]) > isovalue2 then
			local px, py = interpolatePoint(corners[4][1], corners[4][2], values[4], corners[3][1], corners[3][2], values[3], isovalue2)
			table.insert(points, px)
			table.insert(points, py)
		end
		if math.min(values[4], values[3]) < isovalue1 and math.max(values[4], values[3]) > isovalue1 then
			local px, py = interpolatePoint(corners[4][1], corners[4][2], values[4], corners[3][1], corners[3][2], values[3], isovalue1)
			table.insert(points, px)
			table.insert(points, py)
		end
	end

	if    values[4] > math.min (isovalue1, isovalue2)
		and values[4] < math.max (isovalue1, isovalue2) then 
		-- adding bottom left vertice
		local px, py = corners[4][1], corners[4][2]
		table.insert(points, px)
		table.insert(points, py)
	end
	
	
	if values[1] > values[4] then
		if math.min(values[1], values[4]) < isovalue1 and math.max(values[1], values[4]) >= isovalue1 then
			local px, py = interpolatePoint(corners[1][1], corners[1][2], values[1], corners[4][1], corners[4][2], values[4], isovalue1)
			table.insert(points, px)
			table.insert(points, py)
		end
		if math.min(values[1], values[4]) < isovalue2 and math.max(values[1], values[4]) >= isovalue2 then
			local px, py = interpolatePoint(corners[1][1], corners[1][2], values[1], corners[4][1], corners[4][2], values[4], isovalue2)
			table.insert(points, px)
			table.insert(points, py)
		end
	else -- other direction
		if math.min(values[1], values[4]) < isovalue2 and math.max(values[1], values[4]) >= isovalue2 then
			local px, py = interpolatePoint(corners[1][1], corners[1][2], values[1], corners[4][1], corners[4][2], values[4], isovalue2)
			table.insert(points, px)
			table.insert(points, py)
		end
	if math.min(values[1], values[4]) < isovalue1 and math.max(values[1], values[4]) >= isovalue1 then
			local px, py = interpolatePoint(corners[1][1], corners[1][2], values[1], corners[4][1], corners[4][2], values[4], isovalue1)
			table.insert(points, px)
			table.insert(points, py)
		end
	end
	

	if #points < 6 then
		return nil
	else
		return points
	end
end


function createPolygons (noiseMap, width, height)
	local polygons = {}
	for x = 1, width do
		for y = 1, height do
			local polygon = createIsopolygon (noiseMap, x, y, isovalue, isovalue+0.05)
			table.insert(polygons, polygon)
		end
	end
	return polygons
end


return iso